@@ -25,8 +25,8 @@ import time
 import psutil
 import psutil._common
 from psutil import LINUX
+from psutil import MACOS
 from psutil import OPENBSD
-from psutil import OSX
 from psutil import POSIX
 from psutil import SUNOS
 from psutil import WINDOWS
@@ -38,7 +38,6 @@ from psutil.tests import HAS_CPU_FREQ
 from psutil.tests import HAS_ENVIRON
 from psutil.tests import HAS_IONICE
 from psutil.tests import HAS_MEMORY_MAPS
-from psutil.tests import HAS_NUM_CTX_SWITCHES
 from psutil.tests import HAS_PROC_CPU_NUM
 from psutil.tests import HAS_PROC_IO_COUNTERS
 from psutil.tests import HAS_RLIMIT
@@ -177,7 +176,7 @@ class TestMemLeak(unittest.TestCase):
     def _get_mem():
         # By using USS memory it seems it's less likely to bump
         # into false positives.
-        if LINUX or WINDOWS or OSX:
+        if LINUX or WINDOWS or MACOS:
             return thisproc.memory_full_info().uss
         else:
             return thisproc.memory_info().rss
@@ -289,7 +288,6 @@ class TestProcessObjectLeaks(TestMemLeak):
         self.execute(self.proc.num_fds)
 
     @skip_if_linux()
-    @unittest.skipIf(not HAS_NUM_CTX_SWITCHES, "not supported")
     def test_num_ctx_switches(self):
         self.execute(self.proc.num_ctx_switches)
 
@@ -346,8 +344,8 @@ class TestProcessObjectLeaks(TestMemLeak):
         with open(TESTFN, 'w'):
             self.execute(self.proc.open_files)
 
-    # OSX implementation is unbelievably slow
-    @unittest.skipIf(OSX, "too slow on OSX")
+    # MACOS implementation is unbelievably slow
+    @unittest.skipIf(MACOS, "too slow on MACOS")
     @unittest.skipIf(not HAS_MEMORY_MAPS, "not supported")
     @skip_if_linux()
     def test_memory_maps(self):
@@ -526,13 +524,14 @@ class TestModuleFunctionsLeaks(TestMemLeak):
 
     # --- net
 
+    @unittest.skipIf(TRAVIS and MACOS, "false positive on travis")
     @skip_if_linux()
     def test_net_io_counters(self):
         self.execute(psutil.net_io_counters, nowrap=False)
 
     @unittest.skipIf(LINUX,
                      "worthless on Linux (pure python)")
-    @unittest.skipIf(OSX and os.getuid() != 0, "need root access")
+    @unittest.skipIf(MACOS and os.getuid() != 0, "need root access")
     def test_net_connections(self):
         with create_sockets():
             self.execute(psutil.net_connections)
